//
// ============ NewtonGameDynamics Integration (1.3.2010) ==========
//
// Main file
//

#include "main.hpp"
#include "common.hpp"

/*
 * Global members
 */

// SoftPixel Engine objects
SoftPixelDevice* spDevice       = 0;

scene::SceneManager* spSmngr    = 0;
video::VideoDriver* spDriver    = 0;
UserControl* spControl          = 0;

scene::Camera* cam  = 0;
scene::Light* light = 0;

scene::Entity* MeshWorld    = 0;

video::Texture* CrateTex[3] = { 0 };

const dim::vector3df DefaultInGameGravity(0, -25, 0);
dim::vector3df InGameGravity(DefaultInGameGravity);

// Newton objets
NewtonWorld* nWorld     = 0;

PhysicsSystem* PhysX    = 0;


/*
 * Main function
 */

int main(void)
{
    const s32 ScrWidth  = 800;
    const s32 ScrHeight = 600;
    
    spDevice    = createDevice(
        ChooseRenderer(), dim::size2di(ScrWidth, ScrHeight), 32, "SoftPixel Engine: \"NewtonGameDynamics\" Integration"
    );
    
    spControl   = spDevice->getUserControl();
    spSmngr     = spDevice->getSceneManager();
    spDriver    = spDevice->getVideoDriver();
    
    CreateScene();
    
    while (spDevice->updateEvent() && !spControl->keyDown(KEY_ESCAPE))
    {
        spDriver->clearBuffers();
        
        UpdateScene();
        
        spSmngr->renderScene();
        
        spDriver->flipBuffers();
    }
    
    ClearScene();
    
    return 0;
}


/*
 * Sub routine functions
 */

void CreateScene()
{
    // Configure the device
    spControl->setCursorVisible(false);
    
    math::SeedRandom();
    
    // Create camera & embient light
    cam = spSmngr->createCamera();
    cam->setRange(0.1, 1000.0);
    cam->setPosition(dim::vector3df(0, 8, -12));
    
    light = spSmngr->createLight();
    light->setRotation(dim::vector3df(70, 25, 0));
    
    spSmngr->setLighting();
    
    // Create the world mesh
    MeshWorld = spSmngr->loadScene(
        "q3map/maps/aepyra.bsp",
        "q3map/",
        scene::SCENE_AUTODETECT,
        5 // Bezier patch tesselation (default is 7, but for physics simulation we use less)
    );
    
    MeshWorld->setLighting(false);
    
    // Load three crate textures
    CrateTex[0] = spDriver->loadTexture("objects/crate1.jpg");
    CrateTex[1] = spDriver->loadTexture("objects/crate2.jpg");
    CrateTex[2] = spDriver->loadTexture("objects/crate3.jpg");
    
    // Create physics system with all its objects
    PhysicsObject* Obj = 0;
    
    PhysX = new PhysicsSystem();
    
    PhysX->CreatePhysicsObject(PHYSICS_ENVIRONMENT, MeshWorld);
    
    for (s32 x = -10, z; x < 10; x += 3)
    {
        for (z = -5; z < 5; z += 3)
        {
            Obj = PhysX->CreatePhysicsObject(PHYSICS_CUBE);
            Obj->SetPosition(dim::vector3df(-x, 10, z));
        }
    }
}

void UpdateScene()
{
    // Camera movement
    tool::moveCameraFree(cam, 0.15);
    
    // User input controls
    if (spControl->mouseHit(MOUSE_LEFT))
    {
        PhysicsObject* Obj = PhysX->CreatePhysicsObject((EPhysicsObjectTypes)math::Rnd(PHYSICS_CYLINDER));
        Obj->SetPosition(cam->getPosition());
        Obj->SetRotation(cam->getRotationMatrix());
        Obj->AddVelocity(cam->getDirection() * 15);
    }
    
    if (spControl->mouseHit(MOUSE_RIGHT))
    {
        scene::EEntityModels ModelType = scene::ENTITY_CUBE;
        f32 Size = 1.0f;
        
        switch (math::Rnd(3))
        {
            case 0:
                ModelType = scene::ENTITY_TEAPOT; break;
            case 1:
                ModelType = scene::ENTITY_ICOSAHEDRON, Size = 2.0f; break;
            case 2:
                ModelType = scene::ENTITY_SPIRAL; break;
            case 3:
                ModelType = scene::ENTITY_TORUS; break;
        }
        
        scene::Entity* Mesh = spSmngr->createModel(ModelType);
        Mesh->meshTransform(Size);
        Mesh->setColor(video::color(math::Rnd(255), math::Rnd(255), math::Rnd(255)));
        
        PhysicsObject* Obj = PhysX->CreatePhysicsObject(PHYSICS_POLYGON, Mesh);
        Obj->SetPosition(cam->getPosition());
        Obj->SetRotation(cam->getRotationMatrix());
        Obj->AddVelocity(cam->getDirection() * 15);
    }
    
    if (spControl->keyHit(KEY_SPACE))
    {
        static bool EnableGravity = true;
        EnableGravity = !EnableGravity;
        
        InGameGravity = (EnableGravity ? DefaultInGameGravity : 0.0f);
    }
    
    // Update physics simulation
    const f32 fps = spDevice->getFPS();
    
    NewtonUpdate(nWorld, 1.0f / fps);
}

void ClearScene()
{
    // Delete the physics system
    spDeletePointer(PhysX);
    
    // Delete the SoftPixel Device with all its objecst
    deleteDevice();
}


/*
 * Static functions
 */

static void _cdecl SetMeshTransformEvent(const NewtonBody* nBody, const f32* Matrix, s32 ThreadIndex)
{
    PhysicsObject* Obj = (PhysicsObject*)NewtonBodyGetUserData(nBody);
    
    if (Obj && Obj->GetMesh())
    {
        // Set the object's transformation matrix computed by the physics engine
        dim::matrix4f Mat;
        memcpy(&Mat[0], Matrix, sizeof(f32)*16);
        
        Mat.scale(Obj->GetMesh()->getScale());
        Obj->GetMesh()->setMatrix(Mat);
    }
}

static void _cdecl ApplyForceAndTorqueEvent(const NewtonBody* nBody, dFloat TimeStep, s32 ThreadIndex)
{
    PhysicsObject* Obj = (PhysicsObject*)NewtonBodyGetUserData(nBody);
    
    if (!Obj)
        return;
    
    // Get the mass information
    f32 Mass;
    f32 Ixx, Iyy, Izz;
    
    NewtonBodyGetMassMatrix(nBody, &Mass, &Ixx, &Iyy, &Izz);
    
    // Set the gravity force (can point in any direction)
    const dim::vector3df Dir = InGameGravity * Mass;
    
    NewtonBodySetForce(nBody, &Dir.X);
}


/*
 * PhysicsObject class
 */

PhysicsObject::PhysicsObject(const EPhysicsObjectTypes Type, const dim::vector3df &Scale)
    : nCollision_(0), nBody_(0), Mesh_(0)
{
    CreateDynamic(Type, Scale);
}
PhysicsObject::PhysicsObject(const EPhysicsObjectTypes Type, scene::Entity* Mesh)
    : nCollision_(0), nBody_(0), Mesh_(Mesh)
{
    if (Type == PHYSICS_ENVIRONMENT)
        CreateStatic();
    else
        CreateDynamic(Type, 1.0f);
}
PhysicsObject::~PhysicsObject()
{
    if (nCollision_)
        NewtonReleaseCollision(nWorld, nCollision_);
    if (nBody_)
        NewtonDestroyBody(nWorld, nBody_);
    if (Mesh_)
        spSmngr->deleteNode(Mesh_);
}

void PhysicsObject::SetPosition(const dim::vector3df &Position)
{
    dim::matrix4f Mat;
    NewtonBodyGetMatrix(nBody_, &Mat[0]);
    Mat.setPosition(Position);
    NewtonBodySetMatrix(nBody_, &Mat[0]);
}
dim::vector3df PhysicsObject::GetPosition() const
{
    dim::matrix4f Mat;
    NewtonBodyGetMatrix(nBody_, &Mat[0]);
    return Mat.getPosition();
}

void PhysicsObject::SetRotation(const dim::matrix4f &Rotation)
{
    dim::matrix4f Mat;
    Mat.translate(GetPosition());
    Mat *= Rotation;
    NewtonBodySetMatrix(nBody_, &Mat[0]);
}
dim::matrix4f PhysicsObject::GetRotation() const
{
    dim::matrix4f Mat;
    NewtonBodyGetMatrix(nBody_, &Mat[0]);
    return Mat.getRotationMatrix();
}

void PhysicsObject::AddForce(const dim::vector3df &Direction)
{
    NewtonBodyAddForce(nBody_, &Direction.X);
}
void PhysicsObject::SetForce(const dim::vector3df &Direction)
{
    NewtonBodySetForce(nBody_, &Direction.X);
}
dim::vector3df PhysicsObject::GetForce() const
{
    dim::vector3df Direction;
    NewtonBodyGetForce(nBody_, &Direction.X);
    return Direction;
}

void PhysicsObject::SetVelocity(const dim::vector3df &Velocity)
{
    NewtonBodySetVelocity(nBody_, &Velocity.X);
}
void PhysicsObject::AddVelocity(const dim::vector3df &Velocity)
{
    SetVelocity(GetVelocity() + Velocity);
}
dim::vector3df PhysicsObject::GetVelocity() const
{
    dim::vector3df Velocity;
    NewtonBodyGetVelocity(nBody_, &Velocity.X);
    return Velocity;
}

void PhysicsObject::AddImpulse(const dim::vector3df &Position, const dim::vector3df &Direction)
{
    NewtonBodyAddImpulse(nBody_, &Direction.X, &Position.X);
}

void PhysicsObject::CreateDynamic(const EPhysicsObjectTypes Type, const dim::vector3df &Scale)
{
    // Create the special physics object type
    switch (Type)
    {
        case PHYSICS_CUBE:
        {
            nCollision_ = NewtonCreateBox(nWorld, Scale.X, Scale.Y, Scale.Z, 0, 0);
            Mesh_       = spSmngr->createModel(scene::ENTITY_CUBE);
        }
        break;
        
        case PHYSICS_CONE:
        {
            nCollision_ = NewtonCreateCone(nWorld, Scale.Y/2, Scale.X, 0, 0);
            Mesh_       = spSmngr->createModel(scene::ENTITY_CONE);
            Mesh_->meshTurn(dim::vector3df(0, 0, -90));
        }
        break;
        
        case PHYSICS_CYLINDER:
        {
            nCollision_ = NewtonCreateCylinder(nWorld, Scale.Y/2, Scale.X*1.5, 0, 0);
            Mesh_       = spSmngr->createModel(scene::ENTITY_CYLINDER);
            Mesh_->meshTurn(dim::vector3df(0, 0, 90));
            Mesh_->meshTransform(dim::vector3df(1.5, 1, 1));
        }
        break;
        
        case PHYSICS_SPHERE:
        {
            nCollision_ = NewtonCreateSphere(nWorld, Scale.X/2, Scale.Y/2, Scale.Z/2, 0, 0);
            Mesh_       = spSmngr->createModel(scene::ENTITY_SPHERE);
        }
        break;
        
        case PHYSICS_POLYGON:
        {
            const s32 VertexCount = Mesh_->getVertexCount();
            
            f32* pVertices = new f32[VertexCount*3];
            
            u32 s, i, j = 0;
            
            for (s = 0; s < Mesh_->getSurfaceCount(); ++s)
            {
                Mesh_->selectSurface(s);
                
                for (i = 0; i < Mesh_->getVertexCount(s); ++i)
                {
                    pVertices[j++] = Mesh_->getVertexCoord(i).X;
                    pVertices[j++] = Mesh_->getVertexCoord(i).Y;
                    pVertices[j++] = Mesh_->getVertexCoord(i).Z;
                }
            }
            
            nCollision_ = NewtonCreateConvexHull(nWorld, VertexCount, pVertices, sizeof(f32)*3, 0.0f, 0, 0);
            
            delete [] pVertices;
        }
        break;
    }
    
    if (Type != PHYSICS_POLYGON)
        Mesh_->addTexture(CrateTex[math::Rnd(2)]);
    
    // Create and configure the physics body
    nBody_ = NewtonCreateBody(nWorld, nCollision_);
    
    Mesh_->meshTransform(Scale);
    
    // Mass and user data
    static const f32 DefaultMass = 35.0f;
    
    NewtonBodySetUserData(nBody_, this);
    NewtonBodySetMassMatrix(nBody_, DefaultMass, 15.0f, 15.0f, 15.0f);
    
    // No auto freeze/sleep
    NewtonBodySetFreezeState(nBody_, 0);
    NewtonBodySetAutoSleep(nBody_, 0);
    
    // Newton callback procedures
    NewtonBodySetTransformCallback(nBody_, SetMeshTransformEvent);
    NewtonBodySetForceAndTorqueCallback(nBody_, ApplyForceAndTorqueEvent);
}

void PhysicsObject::CreateStatic()
{
    f32 vArray[9];
    u32 Indices[3];
    dim::vector3df Pos[3];
    
    u32 s, i, j;
    
    // Create collision tree and fill the triangle data
    nCollision_ = NewtonCreateTreeCollision(nWorld, 0);
    NewtonTreeCollisionBeginBuild(nCollision_);
    
    for (s = 0; s < Mesh_->getSurfaceCount(); ++s)
    {
        Mesh_->selectSurface(s);
        
        for (i = 0; i < Mesh_->getTriangleCount(s); ++i)
        {
            Mesh_->getTriangleIndices(i, Indices);
            
            for (j = 0; j < 3; ++j)
            {
                Pos[j] = Mesh_->getVertexCoord(Indices[j]);
                
                vArray[j*3+0] = Pos[j].X;
                vArray[j*3+1] = Pos[j].Y;
                vArray[j*3+2] = Pos[j].Z;
            }
            
            NewtonTreeCollisionAddFace(nCollision_, 3, vArray, sizeof(dim::vector3df), 0);
        }
    }
    
    // Finish collision tree creation
    NewtonTreeCollisionEndBuild(nCollision_, 1);
    nBody_ = NewtonCreateBody(nWorld, nCollision_);
    
    f32 BoxP0[3];
    f32 BoxP1[3];
    f32 pMat[4][4];
    
    NewtonBodyGetMatrix(nBody_, &pMat[0][0]);
    NewtonCollisionCalculateAABB(nCollision_, &pMat[0][0], &BoxP0[0], &BoxP1[0]);
}


/*
 * PhysicsSystem class
 */

PhysicsSystem::PhysicsSystem()
{
    // Create the physics environment
    nWorld = NewtonCreate();
    
    // Create a physics material (in this example only one)
    s32 MaterialID = NewtonMaterialGetDefaultGroupID(nWorld);
    
    // Configure material (friction, elasticity, softness)
    NewtonMaterialSetDefaultFriction(nWorld, MaterialID, MaterialID, 0.8f, 0.4f);
    NewtonMaterialSetDefaultElasticity(nWorld, MaterialID, MaterialID, 0.3f);
    NewtonMaterialSetDefaultSoftness(nWorld, MaterialID, MaterialID, 0.05f);
}
PhysicsSystem::~PhysicsSystem()
{
    // Delete each physics object
    for (list<PhysicsObject*>::iterator it = PhysicsObjects_.begin(); it != PhysicsObjects_.end(); ++it)
        spDeletePointer(*it);
    
    // Destroy physics environment
    NewtonDestroy(nWorld);
}

PhysicsObject* PhysicsSystem::CreatePhysicsObject(const EPhysicsObjectTypes Type)
{
    PhysicsObject* Obj = new PhysicsObject(Type);
    PhysicsObjects_.push_back(Obj);
    return Obj;
}
PhysicsObject* PhysicsSystem::CreatePhysicsObject(const EPhysicsObjectTypes Type, scene::Entity* Mesh)
{
    PhysicsObject* Obj = new PhysicsObject(Type, Mesh);
    PhysicsObjects_.push_back(Obj);
    return Obj;
}
void PhysicsSystem::RemovePhysicsObject(PhysicsObject* &Object)
{
    for (list<PhysicsObject*>::iterator it = PhysicsObjects_.begin(); it != PhysicsObjects_.end(); ++it)
    {
        if (*it == Object)
        {
            spDeletePointer(Object);
            PhysicsObjects_.erase(it);
            break;
        }
    }
}





// ================================================================
